# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.abspath('..'))
from druzhba.main import _get_parser
from druzhba import __version__


gen_dir = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    "_gen",
)
if not os.path.exists(gen_dir):
    os.makedirs(gen_dir)

cli_help = _get_parser().format_help()
cli_help = cli_help.replace("sphinx-build", "druzhba")  # argparse grabs current process name
cli_help = '\n'.join(['  ' + line for line in cli_help.split('\n')])
with open(os.path.join(gen_dir, 'usage.rst'), 'w') as f:
    f.write('.. code-block:: text\n\n')
    f.write(cli_help)
    f.write('\n')


def pretty_version(release_str: str):
    """
    >>> pretty_version('0.2.0-rc1')
    0.2
    """
    return '.'.join(release_str.split('.')[:2])


# -- Project information -----------------------------------------------------

project = 'Druzhba'
copyright = '2020, Seatgeek and Contributors'
author = 'Seatgeek and Contributors'

# The full version, including alpha/beta/rc tags
release = __version__
version = pretty_version(release)

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = []

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', 'release.rst']

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'alabaster'
html_theme_options = {
    'show_powered_by': False,
    'logo': 'Druzhba_Logo_Tight.jpg',
    'description': 'A friendly data pipeline framework',
}

html_sidebars = {
    "**": ['about.html', 'navigation.html', 'searchbox.html'],
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']
html_css_files = [
    'https://fonts.googleapis.com/css?family=Roboto',
    'https://fonts.googleapis.com/css?family=Roboto+Mono',
]


rst_prolog = """
.. |example-link| raw:: html

    <a href="https://github.com/seatgeek/druzhba/tree/{version}/test/integration/config">examples</a>
""".format(version=version)




import datetime
import decimal
import json
import unicodedata
import uuid

from fastavro._write_py import WRITERS
from fastavro.write import Writer


def _avro_format(inp):
    if isinstance(inp, uuid.UUID):
        return str(inp)
    if isinstance(inp, (dict, list, tuple)):
        return json.dumps(inp)
    if type(inp) == datetime.datetime:
        # datetime is a subclass of date but the api for isoformat is different
        # so we can't use isinstance here
        return inp.isoformat(" ")
    if isinstance(inp, datetime.date):
        return inp.isoformat()
    if isinstance(inp, datetime.timedelta):
        return (datetime.datetime.min + inp).time().isoformat()
    if isinstance(inp, decimal.Decimal):
        return str(inp)
    if isinstance(inp, str):
        return (
            unicodedata.normalize("NFKD", inp).encode("ascii", "ignore").decode("ascii")
        )
    return inp


def _format_row(inp):
    return {k: _avro_format(v) for k, v in inp.items()}


def write_avro_file(f, results_iter, fields, table, max_size=100 * 1024 ** 2):
    """Takes a database result set (list of dicts) and writes an avro file
    up to a particular size. If the schema 'name' is the same as an Avro data
    type (WRITERS.keys()) everything will break for no apparent reason. 'name'
    isn't even really used.

    Returns complete, row_count

    complete is true if the entire results_iter has been drained -- false if
    there are more records to be processed.

    row_count is the number of items written

    max_size is limit at which we should start writing another file.
    """

    if table in WRITERS:
        table += "zzz"

    schema = {"type": "record", "name": table, "fields": fields}

    writer = Writer(f, schema)

    row_count = 0
    complete = False

    try:
        # writer.io buffers before writing
        while f.tell() + writer.io.tell() < max_size:
            writer.write(_format_row(next(results_iter)))
            row_count += 1
    except StopIteration:
        complete = True
    finally:
        writer.flush()

    return complete, row_count



from contextlib import closing

import pymssql

from druzhba.config import CONFIG_DIR
from druzhba.table import TableConfig, load_query


class MSSQLTableConfig(TableConfig):
    """TableConfig Subclass for MSSQL tables. This class will have all
    methods that are specific to MSSQL, and handle the extracting and
    transforming of the data.

    Attributes
    ----------
    see TableConfig
    """

    database_type = "mssql"
    avro_type_map = {
        # Lint escapes here are because pylint cant introspect what appears to be
        # Java code in pymssql module
        "string": {
            pymssql.STRING.value,  # pylint: disable=no-member
            pymssql.DATETIME.value,  # pylint: disable=no-member
            pymssql.BINARY.value,  # pylint: disable=no-member
        },
        "int": {},  # prefer long to int
        "long": {pymssql.NUMBER.value},  # pylint: disable=no-member
        "double": {},
        "boolean": {},
        "decimal": {pymssql.DECIMAL.value},  # pylint: disable=no-member
    }

    def _load_new_index_value(self):
        raise NotImplementedError(
            "auto generated index queries not yet supported for MSSQL"
        )

    @property
    def connection_vars(self):
        return {
            "server": self.db_host,
            "user": self.db_user,
            "password": self.db_password,
            "database": self.db_name,
            "port": self.db_port,
            "login_timeout": 10,
            "charset": "UTF-8",  # default, but be explicit
        }

    def get_sql_description(self, sql):
        with closing(pymssql.connect(**self.connection_vars)) as conn:
            with closing(conn.cursor(as_dict=True)) as cursor:
                cursor.execute(sql)
                return ((col[0], col[1]) for col in cursor.description)

    def query_to_redshift_create_table(self, sql, table_name):
        if self.schema_file:
            query = load_query(self.schema_file, CONFIG_DIR)
            create_table = query.rstrip("; \n")
            create_table += self.create_table_keys()
            return create_table
        else:
            msg = "auto generated queries not yet supported for MSSQL"
            raise NotImplementedError(msg)

    def query(self, sql):
        """Handles the opening and closing of connections for querying the
        source database

        Parameters
        ----------
        sql : str
            preformatted SQL query

        Returns
        -------
        list of dicts (generator)
            Returns the full database result. If a query returns no
            results, this returns an empty list.
        """
        self.logger.debug("Running query: %s", sql)
        with closing(pymssql.connect(**self.connection_vars)) as conn:
            with closing(conn.cursor(as_dict=True)) as cursor:
                cursor.execute(sql)
                for dict_row in cursor:
                    yield dict_row

    def _get_query_sql(self):
        if self.query_file:
            return self.get_query_from_file()
        else:
            msg = "auto generated queries not yet supported for MSSQL"
            raise NotImplementedError(msg)



import datetime
import json
import logging
import os
import time
from collections import namedtuple
from io import BytesIO

from boto3.s3.transfer import TransferConfig
from boto3.session import Session
from jinja2 import Environment, FileSystemLoader, StrictUndefined, select_autoescape

from druzhba.avro import write_avro_file
from druzhba.config import CONFIG_DIR
from druzhba.redshift import (
    generate_copy_query,
    generate_count_query,
    generate_create_table_like_query,
    generate_drop_exists_query,
    generate_drop_query,
    generate_insert_all_query,
    generate_lock_query,
    generate_rename_query,
    get_redshift,
)


def load_query(query, query_dir):
    with open(os.path.join(query_dir, query)) as f:
        query_file_contents = f.read()
    return query_file_contents


def _s3_url(bucket, key):
    return "s3://{}/{}".format(bucket, key)


class ConfigurationError(Exception):
    def __init__(self, msg, table):
        message = "Conflicting configuration for table {}: {}".format(table, msg)
        super(ConfigurationError, self).__init__(message)


class InvalidSchemaError(Exception):
    pass


class TableStateError(Exception):
    pass


class MigrationError(Exception):
    pass


class Permissions(namedtuple("Permissions", ["name", "is_group", "grants", "owner"])):
    all_str = "arwdRxt"
    all_grants = "ALL PRIVILEGES"
    char_to_grant = {
        "r": "SELECT",  # read
        "w": "UPDATE",  # write
        "a": "INSERT",  # append
        "d": "DELETE",
        "D": "TRUNCATE",
        "x": "REFERENCES",
        "t": "TRIGGER",
        "R": "RULE",  # not documented, apparently
    }

    @classmethod
    def parse(cls, raw_permissions):
        """

        Parameters
        ----------
        raw_permissions : str
            Like {user_name=arwdRxt/owner_name,"group group_name=r/owner_name"}

        Returns
        -------
        List of Permissions, or None if they could not be parsed
        """
        if not raw_permissions:
            return []
        elif raw_permissions[0] != "{" or raw_permissions[-1] != "}":
            return None
        else:
            output = []
            for p in raw_permissions.strip("{}").split(","):
                user, permission = p.strip('"').split("=")
                levels, owner = permission.split("/")
                if user == "":
                    name, is_group = "PUBLIC", True
                elif user.startswith("group "):
                    name, is_group = user[6:], True
                else:
                    name, is_group = user, False

                # A following * represents WITH GRANT OPTION - ignore
                levels_stripped = levels.replace("*", "")
                if levels_stripped == cls.all_str:
                    grants = [cls.all_grants]
                else:
                    grants = [cls.char_to_grant[c] for c in levels_stripped]

                output.append(cls(name, is_group, grants, owner))
            return output


class TableConfig(object):
    """Base class for a specific table. This class will have all methods
    that are engine agnostic--methods that only act with the host server
    or the data warehouse. All methods that interact with a specific engine
    should raise a NotImplementedError here and be overwritten with a
    subclass.

    Parameters
    ----------
    database_alias : str
        Config file name of the database
    db_connection_params : db.ConnectionParams
        DB connection parameters derived from a parsed connection string
    destination_table_name : str
        The name of the table where data should be loaded in the data
        warehouse
    destination_schema_name : str
        The name of the schema where data should be loaded in the data
        warehouse. Note : this should be "public" unless there's a good
        reason to segregate data.
    source_table_name : str
        name of the table in the source database. If a query_file is provided,
        this is purely for logging and monitoring
    query_file : str, optional
        path of a fully qualified SQL file that contains all the logic
        needed to extract a table.
    columns_to_drop : list of str, optional
        Defined by the YAML file, a list of columns that should not be
        imported into the data warehouse
    distribution_key : str, optional
        destination column name to use as the Redshift distkey
    sort_keys : list of str, optional
        destination column names to use as the Redshift sortkeys
    append_only : bool, optional
        Indicates that rows should only be inserted to this table and
        never updated or deleted. If True, primary_key has no effect.
        Default: False
    primary_key : str or list(str), optional
        Columns used to match records when updating the destination table. If
        not provided, primary keys are inferred from the source table. Has no
        effect if append_only is True
        Default: None
    index_column : str, optional
        Column used to identify new or updated rows in the source table.
        Persisted in the index table.
    index_sql : str, optional
        Custom SQL to be run against the source DB to find the current
        max index. Standard templating for the table applies and Druzhba expects
        the index to be returned in a column called `index_value`. Overrides
        index_column.
    truncate_file : str, optional
        Path to a fully qualified SQL file that contains logic to truncate
        the table when full-refreshing. Required to pass --full-refresh if a
        table defined is defined by a query_file.
    full_refresh : boolean, optional
        flag that forces a full deletion of a table prior to loading new data,
        rather than deleting only on matched PKs.  Setting True will conflict
        with index_column.
    rebuild : boolean, optional
        flag to rebuild the table completely. Implies full_refresh.
        Incompatible with query_file and certain conditions per-database-driver.
    type_map : dict, optional
        override type conversion from the source DB to redshift. This is
        set by YAML, and is an empty list of no configuration is provided.
        ex: {
               'tinyint(1)': 'smallint',
               'char(35)': 'varchar(70)',
               'bigint(20) unsigned': 'bigint'
            }

    Attributes
    ----------
    columns : list of str
        All columns read from the source table
    foreign_keys : list of str
        generated from the `create table` syntax, a list of foreign key
        relationships to create for a table after all tables are created
    comments : list of str
        generated from the `create table` syntax, a list of COMMENT ON
        commands to add comments to table columns
    pks : list of str
        generated from the `create table` syntax, a list of source column
        names that define the PK
    key_name : str
        Once the dump has happened, the file is uploaded to s3 at this
        key location
    old_index_value : str
        Used in the `where` clause, this is the most recent index value for
        a given table currently in the data warehouse
    new_index_value : str
        Used in the `where` clause, this is the max index value for a given
        table currently in the source database.
    data : dict
        Read from table definition in the yaml file to supply data to the Jinja
        templating
    db_template_data : dict
        Read from db definition in the yaml file to supply data to the Jinja
        templating

    Notes
    -----
    All parameters are also set as attributes
    """

    DESTINATION_TABLE_OK = "ok"
    DESTINATION_TABLE_REBUILD = "rebuild"
    DESTINATION_TABLE_DNE = "non-existent"
    DESTINATION_TABLE_INCORRECT = "incorrect"

    max_file_size = 100 * 1024 ** 2

    def __init__(
        self,
        database_alias,
        db_connection_params,
        destination_table_name,
        destination_schema_name,
        source_table_name,
        index_schema,
        index_table,
        query_file=None,
        distribution_key=None,
        sort_keys=None,
        index_column=None,
        index_sql=None,
        truncate_file=None,
        columns_to_drop=None,
        type_map=None,
        primary_key=None,
        not_null_date=False,
        full_refresh=False,
        rebuild=False,
        schema_file=None,
        data=None,
        append_only=False,
        db_template_data=None,
    ):
        self.database_alias = database_alias
        self.db_host = db_connection_params.host
        self.db_port = db_connection_params.port
        self.db_name = db_connection_params.name
        self.db_user = db_connection_params.user
        self.db_password = db_connection_params.password
        self._columns = None
        self.columns_to_drop = columns_to_drop or []
        self.destination_table_name = destination_table_name
        self.destination_schema_name = destination_schema_name
        self.query_file = query_file
        self.schema_file = schema_file
        self.distribution_key = distribution_key
        self.sort_keys = sort_keys
        self.source_table_name = source_table_name
        self.index_column = index_column
        self.index_sql = index_sql
        self.truncate_file = truncate_file
        self.append_only = append_only
        self.full_refresh = full_refresh
        self.rebuild = rebuild
        self.type_map = self._clean_type_map(type_map)
        self.primary_key = (
            [primary_key] if isinstance(primary_key, str) else primary_key
        )
        self.not_null_date = not_null_date
        self.foreign_keys = []
        self.comments = []
        self.pks = []
        self._old_index_value = "notset"
        self._new_index_value = "notset"
        self._destination_table_status = None
        self.table_template_data = data
        self.db_template_data = db_template_data
        self.index_schema = index_schema
        self.index_table = index_table

        self.date_key = datetime.datetime.strftime(
            datetime.datetime.utcnow(), "%Y%m%dT%H%M%S"
        )

        self.row_count = None
        self.upload_size = 0

        self.starttime = None
        self.endtime = None
        self.rows_inserted = None
        self.rows_deleted = None

        self.num_data_files = 0
        self.manifest_mode = False

        self.logger = logging.getLogger(f"druzhba.{database_alias}.{source_table_name}")
        self.s3 = Session().client("s3")

    @classmethod
    def _clean_type_map(cls, type_map):
        if not type_map:
            return {}
        for k, v in type_map.items():
            type_map[k.lower()] = v
        return type_map

    @classmethod
    def validate_yaml_configuration(cls, yaml_config):
        """
        Validate YAML configuration. Note that this can differ slightly from runtime
        config, since full_refresh may be forced even when it would fail these checks.
        """

        table = yaml_config["source_table_name"]
        index_column = yaml_config.get("index_column")
        index_sql = yaml_config.get("index_sql")
        append_only = yaml_config.get("append_only")
        full_refresh = yaml_config.get("full_refresh")
        primary_key = yaml_config.get("primary_key")
        query_file = yaml_config.get("query_file")
        schema_file = yaml_config.get("schema_file")

        has_incremental_index = index_column or index_sql

        if not has_incremental_index and append_only:
            raise ConfigurationError("Append_only without incremental index", table)
        if full_refresh and append_only:
            raise ConfigurationError("Append_only with full_refresh", table)
        elif not has_incremental_index and not full_refresh:
            raise ConfigurationError(
                "Incremental update with no specified index", table
            )
        elif index_column and full_refresh:
            raise ConfigurationError("Full refresh with index_column", table)
        elif index_sql and full_refresh:
            raise ConfigurationError("Full refresh with index_sql", table)
        elif index_sql and index_column:
            raise ConfigurationError("index_sql and index_column", table)
        elif query_file and not primary_key and not append_only and not full_refresh:
            raise ConfigurationError(
                "incremental query_file without primary_key", table
            )
        elif query_file and not os.path.isfile(os.path.join(CONFIG_DIR, query_file)):
            raise ConfigurationError("nonexistent query_file", table)
        elif schema_file and not os.path.isfile(os.path.join(CONFIG_DIR, schema_file)):
            raise ConfigurationError("nonexistent schema_file", table)

    def validate_runtime_configuration(self):
        """
        Validate this instance's configuration state, which can differ from yaml
        configurations allowed by validate_yaml_configuration and runs after
        connecting to the database.
        """
        if self.rebuild and self.truncate_file:
            msg = (
                "Cannot rebuild a table with a truncate_file "
                "because it would not be correct to drop the table."
            )
            raise ConfigurationError(msg, self.source_table_name)
        elif self.rebuild and self.schema_file:
            msg = (
                "Cannot rebuild a table with a schema file, need "
                "support for passing in the table name to create."
            )
            raise ConfigurationError(msg, self.source_table_name)

    @property
    def s3_key_prefix(self):
        return "{}/{}.{}.{}".format(
            get_redshift().s3_config.prefix,
            self.database_alias,
            self.source_table_name,
            self.date_key,
        )

    def single_s3_data_key(self):
        """Returns the S3 path to upload a single avro file to"""
        if self.manifest_mode:
            raise TableStateError(
                "Attempted to treat a manifest upload as a single file"
            )

        return "{}.avro".format(self.s3_key_prefix)

    def manifest_s3_data_key(self):
        if not self.manifest_mode:
            raise TableStateError(
                "Attempted to treat a single file upload as a manifest"
            )

        return "{}.manifest".format(self.s3_key_prefix)

    def numbered_s3_data_key(self, file_num):
        return "{}/{:05d}.avro".format(self.s3_key_prefix, file_num)

    def next_s3_data_file_key(self):
        if self.manifest_mode:
            return self.numbered_s3_data_key(self.num_data_files)
        else:
            return self.single_s3_data_key()

    @property
    def copy_target_key(self):
        if self.manifest_mode:
            return self.manifest_s3_data_key()
        else:
            return self.single_s3_data_key()

    @property
    def copy_target_url(self):
        return _s3_url(get_redshift().s3_config.bucket, self.copy_target_key)

    def data_file_keys(self):
        if self.manifest_mode:
            for fn in range(self.num_data_files):
                yield self.numbered_s3_data_key(fn)
        else:
            yield self.single_s3_data_key()

    @property
    def connection_vars(self):
        raise NotImplementedError

    @property
    def avro_type_map(self):
        raise NotImplementedError

    def get_query_from_file(self):
        env = Environment(
            loader=FileSystemLoader(os.path.join(CONFIG_DIR)),
            autoescape=select_autoescape(["sql"]),
            undefined=StrictUndefined,
        )
        template = env.get_template(self.query_file)
        return template.render(
            db=self.db_template_data,
            table=self.table_template_data,
            run=self.run_template_data,
        )

    def get_sql_description(self, sql):
        raise NotImplementedError

    def get_query_sql(self):
        if self.query_file:
            return self.get_query_from_file()
        else:
            return self._get_query_sql() + self.where_clause()

    def _get_query_sql(self):
        raise NotImplementedError

    @property
    def columns(self):
        if not self._columns:
            self._columns = [
                row[0] for row in self.get_sql_description(self.get_query_sql())
            ]
        return self._columns

    def query(self, sql):
        raise NotImplementedError

    def query_fetchone(self, sql):
        results = self.query(sql)
        return next(results)

    def _check_index_values(self):
        """Returns true if new index is greater than old index and defined"""
        if self.full_refresh:
            if self.index_column or self.index_sql:
                msg = (
                    "Index was found, but %s was forced. "
                    "Old index value will be ignored, but new "
                    "index value will be recorded."
                )
                self.logger.info(msg, "rebuild" if self.rebuild else "full-refresh")
            return True

        if self.append_only:
            return True

        # If there is no previous index, we're fine
        if self.old_index_value is None:
            return True

        # There's an old index but can't load a new value.
        if self.new_index_value is None and self.old_index_value is not None:
            msg = "Index expected but not found. Last value was %s. Dumping full table"
            self.logger.warning(msg, self.old_index_value)
            return False

        try:
            # old_index_value comes in as a unicode new_index_value as the sql
            # type
            if isinstance(self.new_index_value, int):
                is_inverted = int(self.old_index_value) > self.new_index_value
            elif isinstance(self.new_index_value, datetime.datetime):
                old_index_dt = datetime.datetime.strptime(
                    self.old_index_value, "%Y-%m-%d %H:%M:%S.%f"
                )
                is_inverted = old_index_dt > self.new_index_value
            else:
                self.logger.warning(
                    "Unknown type %s for index %s",
                    type(self.new_index_value),
                    self.old_index_value,
                )
                return False
        except (ValueError, TypeError) as ex:
            self.logger.warning("Could not check index: %s", str(ex))
            return False

        if is_inverted:
            self.logger.warning(
                "Index value has decreased for table %s.%s. "
                "May need to do full refresh",
                self.db_name,
                self.source_table_name,
            )
        return not is_inverted

    def row_generator(self):
        sql = self.get_query_sql()
        self._check_index_values()
        self.logger.info("Extracting %s table %s", self.db_name, self.source_table_name)
        self.logger.debug("Running SQL: %s", sql)
        return self.query(sql)

    @property
    def run_template_data(self):
        return {
            "destination_schema_name": self.destination_schema_name,
            "destination_table_name": self.destination_table_name,
            "db_name": self.db_name,
            "source_table_name": self.source_table_name,
            "index_column": self.index_column,
            "new_index_value": self.new_index_value,
            "old_index_value": self.old_index_value,
        }

    def _load_old_index_value(self):
        """Sets and gets the index_value property, retrieved from Redshift

        Returns
        -------
        index_value : variable
            Since index_value can vary from table to table, this can be many
            different types. Most common will be a datetime or int, but
            could also be a date or string. Returns None if no previous index
            value found
        """

        query = f"""
        SELECT index_value
          FROM "{self.index_schema}"."{self.index_table}"
         WHERE datastore_name = %s
           AND database_name = %s
           AND table_name = %s
        ORDER BY created_ts DESC
        LIMIT 1;
        """
        self.logger.debug("Querying Redshift for last updated index")
        with get_redshift().cursor() as cur:
            cur.execute(
                query, (self.database_alias, self.db_name, self.source_table_name)
            )
            index_value = cur.fetchone()

        if index_value is None:
            self.logger.info(
                "No index found. Dumping entire table: %s.", self.source_table_name
            )
            return index_value
        else:
            self.logger.info("Index found: %s", index_value[0])
            return index_value[0]

    @property
    def old_index_value(self):
        # we use 'notset' rather than None because None is a valid output
        if self._old_index_value is "notset":
            self._old_index_value = self._load_old_index_value()
        return self._old_index_value

    def _load_new_index_value(self):
        # Abstract to support DB-specific quoting
        raise NotImplementedError

    @property
    def new_index_value(self):
        # we use 'notset' rather than None because None is a valid output
        if self._new_index_value is "notset":
            if self.index_sql:
                env = Environment(
                    autoescape=select_autoescape(["sql"]), undefined=StrictUndefined
                )
                template = env.from_string(self.index_sql)
                query = template.render(
                    db=self.db_template_data,
                    table=self.table_template_data,
                    # Cannot include the run template data
                    # here because we do not know the index values yet
                )
                self._new_index_value = self.query_fetchone(query)["index_value"]

            elif self.index_column:
                self._new_index_value = self._load_new_index_value()
            else:
                self._new_index_value = None
        if self.query_file and self.index_sql and self._new_index_value is None:
            # Handles a special case where the index_sql query returns no rows
            # but the custom sql file is expecting both old and new index values
            return 0
        return self._new_index_value

    def where_clause(self):
        """Method for filtering get_data, if tables are able to be
        sliced on some index value.

        Returns
        -------
        str
            valid SQL featuring just the WHERE clause
        """
        where_clause = "\nWHERE "
        if not self.index_column or self.full_refresh:
            # If no index_column, there is no where clause. The whole
            # source table is dumped.
            return ""

        if self.new_index_value is None:
            # Either the table is empty or the index_column is all NULL
            return ""

        if self.old_index_value:
            # This should always happen except on the initial load
            where_clause += "{} > '{}' AND ".format(
                self.index_column, self.old_index_value
            )

        where_clause += "{} <= '{}'".format(self.index_column, self.new_index_value)
        return where_clause

    def get_destination_table_columns(self):
        query = """
        SELECT "column"
          FROM pg_table_def
         WHERE schemaname = %s
           AND tablename = %s;
        """

        with get_redshift().cursor() as cur:
            self.set_search_path(cur)
            cur.execute(
                query, (self.destination_schema_name, self.destination_table_name)
            )
            results = cur.fetchall()

        return [x[0] for x in results]

    def get_destination_table_status(self):
        """Queries the data warehouse to determine if the desired
        destination table exists and if so, if it matches the expected
        configuration.

        Returns
        -------
        str
           Representing our plan for what to do with the destination table
           Includes:
            - DESTINATION_TABLE_DNE -> build it if possible
            - DESTINATION_TABLE_REBUILD -> rebuild it
            - DESTINATION_TABLE_OK -> leave it
            - DESTINATION_TABLE_INCORRECT -> error
        """
        dw_columns = set(self.get_destination_table_columns())
        source_columns = set(self.columns)
        expected = source_columns - set(self.columns_to_drop)
        unexpected_dw_columns = dw_columns - expected
        unexpected_source_columns = expected - dw_columns

        if len(dw_columns) == 0:
            self.logger.info("Destination table does not exist.")
            return self.DESTINATION_TABLE_DNE
        elif self.rebuild:
            # We're rebuilding it so we don't care if it's right,
            # so exit before we log any errors
            self.logger.info("Attempting to rebuild destination table.")
            return self.DESTINATION_TABLE_REBUILD
        elif dw_columns == expected:
            return self.DESTINATION_TABLE_OK
        elif len(unexpected_dw_columns) > 0:
            msg = (
                "Columns exist in the warehouse table that are not in "
                "the source: `%s`"
            )
            self.logger.warning(msg, "`, `".join(unexpected_dw_columns))
            return self.DESTINATION_TABLE_INCORRECT
        elif len(unexpected_source_columns) > 0:
            msg = (
                "Columns exist in the source table that are not in the "
                + "warehouse. Skipping column(s): `%s`"
            )
            self.logger.warning(msg, "`, `".join(unexpected_source_columns))

            # Copy from avro will just ignore the extra columns so we can proceed
            return self.DESTINATION_TABLE_OK
        else:
            raise RuntimeError("Unhandled case in get_destination_table_status")

    def query_description_to_avro(self, sql):
        desc = self.get_sql_description(sql)
        fields = []

        for col_desc in desc:
            col_name = col_desc[0]
            schema = {"name": col_name}
            try:
                col_type = col_desc[1].split("(")[0]
            except AttributeError:
                col_type = col_desc[1]
            if col_type in self.avro_type_map["string"]:
                schema["type"] = ["null", "string"]
            elif col_type in self.avro_type_map["int"]:
                schema["type"] = ["null", "int"]
            elif col_type in self.avro_type_map["double"]:
                schema["type"] = ["null", "double"]
            elif col_type in self.avro_type_map["long"]:
                schema["type"] = ["null", "long"]
            elif col_type in self.avro_type_map["boolean"]:
                schema["type"] = ["null", "boolean"]
            elif col_type in self.avro_type_map["decimal"]:
                # fastavro now supports decimal types, but Redshift does not
                schema["type"] = ["null", "string"]
            else:
                self.logger.warning(
                    "unmatched data type for column %s in %s table %s",
                    col_desc[0],
                    self.db_name,
                    self.source_table_name,
                )
                schema["type"] = ["null", "string"]

            fields.append(schema)
        return fields

    def set_last_updated_index(self):
        """Adds a new index to the pipeline_table_index table for updated
        tables
        """
        if self.new_index_value is None:
            return

        query = f"""
        INSERT INTO "{self.index_schema}"."{self.index_table}" VALUES
        (%s, %s, %s, %s)
        """

        if isinstance(self.new_index_value, int):
            new_index_value = str(self.new_index_value)
        elif isinstance(self.new_index_value, datetime.datetime):
            new_index_value = self.new_index_value.strftime("%Y-%m-%d %H:%M:%S.%f")
        else:
            msg = "Don't know how to handle index {} of type {}".format(
                self.new_index_value, str(type(self.new_index_value))
            )
            raise TypeError(msg)

        self.logger.info("Updating index table")
        with get_redshift().cursor() as cur:
            args = (
                self.database_alias,
                self.db_name,
                self.source_table_name,
                new_index_value,
            )
            self.logger.debug(cur.mogrify(query, args))
            cur.execute(query, args)

    def create_table_keys(self, distkey=None, sortkeys=None):
        output = ""
        distkey = distkey or self.distribution_key
        if distkey:
            output += "distkey({})\n".format(distkey)

        sortkeys = sortkeys or self.sort_keys
        if sortkeys:
            output += "compound " if len(sortkeys) > 1 else ""
            output += "sortkey({})\n".format(",".join(sortkeys))
        return output

    def query_to_redshift_create_table(self, sql, table_name):
        raise NotImplementedError

    def check_destination_table_status(self):
        """Get the source table schema, convert it to Redshift compatibility,
        and ensure the table exists as expected in the data warehouse.

        Sets self._destination_table_status if we can proceed, or raises
        if not.

        Raises
        ------
        InvalidSchemaError
            Raised when the target table has columns not recognized in the source
            table, and we're not rebuilding the table. (Unrecognized source
            columns will log but still proceed)
        MigrationError
            Raised when the target table needs to be create or rebuilt but
            cannot be done automatically.
        """
        self.logger.info("Getting CREATE TABLE command")

        self._destination_table_status = self.get_destination_table_status()

        if self._destination_table_status in (
            self.DESTINATION_TABLE_DNE,
            self.DESTINATION_TABLE_REBUILD,
        ):
            self.logger.info("Verifying that the table can be created.")
            try:
                # Only called to see if it raises, the actual table
                # will be created later
                self.query_to_redshift_create_table(
                    self.get_query_sql(), self.destination_table_name
                )
            except NotImplementedError:
                raise MigrationError("Automatic table creation was not implemented for "
                                     "this database, manual migration needed.")
        elif self._destination_table_status == self.DESTINATION_TABLE_INCORRECT:
            raise InvalidSchemaError(
                "Extra columns exist in redshift table. Migration needed"
            )

    def register_extract_monitor(self, starttime, endtime):
        """Adds an entry into the extract monitor for a given extract task

        Parameters
        ----------
        starttime : datetime.datetime
            datetime object generated at the beginning of the data
            extraction
        endtime : datetime.datetime
            datetime object generated at the end of the data extraction
        """

        query = """
        INSERT INTO "public"."table_extract_detail" VALUES (
            %(task_id)s, %(class_name)s, %(task_date_params)s,
            %(task_other_params)s, %(start_dt)s, %(end_dt)s, %(run_time_sec)s,
            %(manifest_path)s, %(data_path)s, %(output_exists)s, %(row_count)s,
            %(upload_size)s, %(exception)s
        );
        """
        args = {
            "task_id": "{}(alias={}, database={}, table={})".format(
                self.__class__.__name__,
                self.database_alias,
                self.db_name,
                self.source_table_name,
            ),
            "class_name": self.__class__.__name__,
            "task_date_params": None,
            "task_other_params": None,
            "start_dt": starttime.replace(microsecond=0),
            "end_dt": endtime.replace(microsecond=0),
            "run_time_sec": (endtime - starttime).total_seconds(),
            "manifest_path": self.copy_target_url,
            "data_path": "s3://{}/{}".format(
                get_redshift().s3_config.bucket, self.s3_key_prefix
            ),
            "output_exists": self.row_count > 0,
            "row_count": self.row_count,
            "upload_size": self.upload_size,
            "exception": None,
        }

        self.logger.info("Inserting record into table_extract_detail")
        with get_redshift().cursor() as cur:
            cur.execute(query, args)

    def register_load_monitor(self):
        """Adds an entry into the load monitor for a given load task

        Parameters
        ----------
        starttime : datetime
            datetime object generated at the beginning of the load
        endtime : datetime
            datetime object generated at the end of the load
        rows_inserted : int
            Total number of rows generated by the Redshift COPY command
        rows_deleted : int
            Count of rows deleted by primary key in the destination table
        load_size : int
            Size in bytes of the file in S3 used in the Redshift COPY
            command
        """

        query = """
            INSERT INTO "public"."table_load_detail" VALUES (
                %(task_id)s, %(class_name)s, %(task_date_params)s, 
                %(task_other_params)s, %(target_table)s, %(start_dt)s,
                %(end_dt)s, %(run_time_sec)s, %(extract_task_update_id)s,
                %(data_path)s, %(manifest_cleaned)s, %(rows_inserted)s,
                %(rows_deleted)s, %(load_size)s, %(exception)s
            );
        """
        task_id = "{}(alias={}, database={}, table={})".format(
            self.__class__.__name__,
            self.database_alias,
            self.db_name,
            self.source_table_name,
        )
        target_table = "{}.{}".format(
            self.destination_schema_name, self.destination_table_name
        )
        args = {
            "task_id": task_id,
            "class_name": self.__class__.__name__,
            "task_date_params": None,
            "task_other_params": None,
            "target_table": target_table,
            "start_dt": self.starttime.replace(microsecond=0),
            "end_dt": self.endtime.replace(microsecond=0),
            "run_time_sec": (self.endtime - self.starttime).total_seconds(),
            "extract_task_update_id": task_id,
            "data_path": self.copy_target_url,
            "manifest_cleaned": False,
            "rows_inserted": self.rows_inserted,
            "rows_deleted": self.rows_deleted,
            "load_size": self.upload_size,
            "exception": None,
        }

        self.logger.info("Inserting record into table_load_detail")
        with get_redshift().cursor() as cur:
            cur.execute(query, args)

    def extract(self):
        """Serializes full db result set and uploads to s3

        The data will be uploaded either as a single file or as a set of files
        with a manifest
        """
        # TODO: Do we not currently execute the extract monitor?
        # starttime = datetime.datetime.utcnow()

        results_schema = self.query_description_to_avro(self.get_query_sql())
        results_iter = self.row_generator()

        done = False
        while not done:
            done = self.avro_to_s3(results_iter, results_schema)

        if self.num_data_files == 0:
            self.logger.info(
                "No data extracted; not uploading to s3 for %s table %s",
                self.db_name,
                self.source_table_name,
            )

        if self.manifest_mode:
            self.write_manifest_file()

        # endtime = datetime.datetime.utcnow()
        # self.register_extract_monitor(starttime, endtime)

    def avro_to_s3(self, results_iter, results_schema):
        """Attempts to serialize a result set to an AVRO file

        returns true if it complete writes the entire result_iter and false
        if there were records remaining when it hit the maximum file size.
        """
        with BytesIO() as f:
            complete, row_count = write_avro_file(
                f,
                results_iter,
                results_schema,
                self.destination_table_name,
                self.max_file_size,
            )

            if self.row_count is None:
                self.row_count = row_count
            else:
                self.row_count += row_count

            self.upload_size += f.tell()

            if not complete:
                self.manifest_mode = True

            if row_count > 0:
                self._upload_s3(
                    f, get_redshift().s3_config.bucket, self.next_s3_data_file_key()
                )
                self.num_data_files += 1

        return complete

    def write_manifest_file(self):
        if not self.manifest_mode:
            raise TableStateError("Cannot write manifest when not in manifest mode")

        entries = [
            {"url": _s3_url(get_redshift().s3_config.bucket, key), "mandatory": True}
            for key in self.data_file_keys()
        ]
        manifest = {"entries": entries}

        with BytesIO() as f:
            f.write(json.dumps(manifest).encode())
            self._upload_s3(
                f, get_redshift().s3_config.bucket, self.manifest_s3_data_key()
            )

    def _upload_s3(self, f, bucket, key):
        """
        Upload a file to this table's s3 key.

        Parameters
        ----------
        f : an open file handle.
        s3_path : string indicating the s3 location to write to
        """

        MB = 1024 ** 2
        s3_config = TransferConfig(multipart_threshold=10 * MB)
        f.seek(0)

        self.logger.info("Writing s3 file %s", _s3_url(bucket, key))

        retries = 3
        retries_remaining = retries
        while retries_remaining > 0:
            try:
                self.s3.upload_fileobj(f, bucket, key, Config=s3_config)
                self.logger.info("Wrote s3 file %s", _s3_url(bucket, key))
                return
            except KeyError:
                # retry on intermittent credential error
                retries_remaining -= 1
                if retries_remaining > 0:
                    time.sleep(3 * (retries - retries_remaining) ** 2)
                else:
                    raise

    def set_search_path(self, cursor):
        """This sets the search_path for a Redshift session.
        The default search_path is "'$user', public"; this replaces it with
        just the destination_schema_name.

        Parameters
        ----------
        cursor : Redshift cursor
            A cursor is passed in rather than generated because we need to
            modify the existing cursor and not create a new one
        """
        query = "SET search_path TO {};".format(self.destination_schema_name)
        self.logger.debug(query)
        cursor.execute(query)

    def get_delete_sql(self):
        if self.full_refresh:
            if self._destination_table_status == self.DESTINATION_TABLE_REBUILD:
                # We'll just drop it
                return ""
            if self.truncate_file:
                env = Environment(
                    loader=FileSystemLoader(os.path.join("datacfg")),
                    autoescape=select_autoescape(["sql"]),
                    undefined=StrictUndefined,
                )
                template = env.get_template(self.truncate_file)
                return template.render(
                    db=self.db_template_data,
                    table=self.table_template_data,
                    run=self.run_template_data,
                )
            else:
                return 'DELETE FROM "{}";'.format(self.destination_table_name)
        elif not self.append_only:
            if self.primary_key:
                # override from db yaml file
                pks = self.primary_key
            else:
                # pk column discovered from the source table
                pks = self.pks

            if not pks and self.index_column:
                raise InvalidSchemaError(
                    "Specifying an index column without primary key would "
                    "result in all records in the existing table being "
                    "deleted. If this is the desired behavior, run with "
                    "--full-refresh instead. If not, check if primary keys"
                    "can be inferred from the upstream database."
                )

            constraints = [
                '"{0}"."{2}" = "{1}"."{2}"'.format(
                    self.staging_table_name, self.destination_table_name, pk
                )
                for pk in pks
            ]
            constraint_string = " AND ".join(constraints)
            return 'DELETE FROM "{}" USING "{}" WHERE {};'.format(
                self.destination_table_name, self.staging_table_name, constraint_string,
            )
        else:
            # Should only land here when append_only
            # in which case we're not deleting
            return None

    def get_grant_sql(self, cursor):
        """ Get SQL statements to restore permissions
        to the staging table after a rebuild. """

        get_permissions_sql = """
        SELECT
            use.usename = CURRENT_USER          AS "owned"
            , c.relacl                          AS "permissions"
        FROM pg_class c
            LEFT JOIN pg_namespace nsp ON c.relnamespace = nsp.oid
            LEFT JOIN pg_user use ON c.relowner = use.usesysid
        WHERE
            c.relkind = 'r'
            AND nsp.nspname = '{schema}'
            AND c.relname = '{table}'
        """.format(
            schema=self.destination_schema_name, table=self.destination_table_name
        )
        cursor.execute(get_permissions_sql)
        permissions_result = cursor.fetchall()

        if len(permissions_result) == 0:
            self.logger.info("No existing permissions found for %s.%s",
                             self.destination_schema_name, self.destination_table_name)
            return None
        elif len(permissions_result) > 1:
            raise MigrationError("Got multiple permissions rows for table")

        is_owner, permissions_str = permissions_result[0]
        if not is_owner:
            raise MigrationError("Can't rebuild target table because it has another owner")

        self.logger.info(
            "Got existing permissions for table to add to %s: %s",
            self.staging_table_name,
            permissions_str,
        )
        permissions = Permissions.parse(permissions_str)
        if permissions is None:
            raise MigrationError(
                "Couldn't parse permissions {} to rebuild target table".format(permissions_str)
            )

        grant_template = "GRANT {grant} ON {table} TO {group}{name};"
        grant_sqls = [
            grant_template.format(
                grant=g,
                table=self.staging_table_name,
                # Should we not restore users?
                group="GROUP " if p.is_group else "",
                name=p.name,
            )
            for p in permissions
            for g in p.grants
        ]
        return "\n".join(grant_sqls)

    @property
    def staging_table_name(self):
        return "{}_{}_staging".format(self.database_alias, self.destination_table_name)

    def load(self):
        """The Load phase of the pipeline. Takes a file in S3 and issues a
        Redshift COPY command to import the data into a staging table. It
        then upserts the data by deleting rows in the destination table
        that match on PK, inserts all rows from the staging table into the
        destination table, and then deletes the staging table.
        """

        self.starttime = datetime.datetime.utcnow()

        # Initializing Data
        delete_clause = self.get_delete_sql()
        staging_table = self.staging_table_name
        destination_table = self.destination_table_name
        is_normal_load = self._destination_table_status == self.DESTINATION_TABLE_OK
        is_rebuild = self._destination_table_status == self.DESTINATION_TABLE_REBUILD
        is_dne = self._destination_table_status == self.DESTINATION_TABLE_DNE

        with get_redshift().cursor() as cur:
            self.set_search_path(cur)

            # If table does not exist, create it
            if is_dne:
                create_table = self.query_to_redshift_create_table(
                    self.get_query_sql(), self.destination_table_name
                )
                cur.execute(create_table)
            elif not is_normal_load and not is_rebuild:
                raise RuntimeError(
                    "Invalid table status in redshift_copy: {}".format(
                        self._destination_table_status
                    )
                )

            # If there is no row updates, just skip copy and return
            if self.row_count == 0:
                return

            cur.execute("BEGIN TRANSACTION;")
            # Lock the table early to avoid deadlocks in many-to-one pipelines.
            query = generate_lock_query(destination_table)
            cur.execute(query)

            query = generate_drop_exists_query(staging_table)
            cur.execute(query)

            if is_rebuild:
                # Build staging table anew and grant it appropriate permissions
                self.logger.info(
                    "Creating staging table to rebuild %s", destination_table
                )
                create_staging_table = self.query_to_redshift_create_table(
                    self.get_query_sql(), staging_table
                )
                permissions_sql = self.get_grant_sql(cur)
                cur.execute(create_staging_table)
                if permissions_sql:
                    self.logger.info(
                        "Copying permissions onto %s:\n%s",
                        staging_table,
                        permissions_sql,
                    )
                    cur.execute(permissions_sql)
            else:
                # If not rebuilding, create staging with LIKE
                self.logger.info("Creating staging table %s", staging_table)
                query = generate_create_table_like_query(
                    staging_table, destination_table
                )
                cur.execute(query)

            # Issuing Copy Command
            self.logger.info("Issuing copy command")
            query = generate_copy_query(
                staging_table,
                self.copy_target_url,
                get_redshift().iam_copy_role,
                self.manifest_mode,
            )
            self.logger.debug(query)
            cur.execute(query)

            # Row delete and count logic
            if is_rebuild or (self.append_only and not self.full_refresh):
                self.rows_deleted = 0
            else:
                cur.execute(delete_clause)
                self.rows_deleted = cur.rowcount

            # Row insert and count logic
            if is_rebuild:
                self.logger.info("Swapping staging table into %s", destination_table)
                # DNE overrides rebuild, so we can assume the table exists
                query = generate_drop_query(destination_table)
                cur.execute(query)
                query = generate_rename_query(staging_table, destination_table)
                cur.execute(query)
                query = generate_count_query(destination_table)
                cur.execute(query)
                self.rows_inserted = cur.fetchall()[0]
            else:
                query = generate_insert_all_query(staging_table, destination_table)
                cur.execute(query)
                self.rows_inserted = cur.rowcount
                query = generate_drop_query(staging_table)
                cur.execute(query)
            cur.execute("END TRANSACTION;")
            self.register_and_cleanup()

    def register_and_cleanup(self):
        # Register in index table.
        self.set_last_updated_index()

        # Register in monitor table
        # self.endtime = datetime.datetime.utcnow()
        # self.register_load_monitor()

        # Clean up S3
        for key in self.data_file_keys():
            self.s3.delete_object(Bucket=get_redshift().s3_config.bucket, Key=key)

        if self.manifest_mode:
            self.s3.delete_object(
                Bucket=get_redshift().s3_config.bucket, Key=self.manifest_s3_data_key()
            )



import datetime
import re
from contextlib import closing

import pymysql.cursors
from pymysql import FIELD_TYPE as ft

from druzhba.config import CONFIG_DIR
from druzhba.table import TableConfig, load_query


class MysqlTypes:
    mysql_type_to_name = {
        ft.TINY: "TINYINT",
        ft.SHORT: "SMALLINT",
        ft.INT24: "MEDIUMINT",
        ft.LONG: "INT",
        ft.LONGLONG: "BIGINT",
        ft.FLOAT: "FLOAT",
        ft.DOUBLE: "DOUBLE",
        ft.NEWDECIMAL: "DECIMAL",
        ft.TIME: "TIME",
        ft.DATE: "DATE",
        ft.DATETIME: "DATETIME",
        ft.TIMESTAMP: "TIMESTAMP",
        ft.STRING: "CHAR",
        ft.VAR_STRING: "VARCHAR",
        ft.TINY_BLOB: "TINYBLOB",
        ft.BLOB: "BLOB",
        ft.MEDIUM_BLOB: "MEDIUMBLOB",
        ft.LONG_BLOB: "LONGBLOB",
        ft.BIT: "BIT",
        ft.JSON: "BLOB",
    }
    numeric_types_always_promote = ["tinyint", "mediumint", "float", "double"]
    numeric_type_promotions = {
        "tinyint": "smallint",
        "mediumint": "integer",
        "smallint": "integer",
        "int": "bigint",
        "bigint": "numeric(65, 0)",
        "float": "real",
        "double": "double precision",
    }
    fixed_types = ["decimal", "numeric"]
    date_and_time_types = {
        "date": "date",
        "time": "varchar(40)",
        "datetime": "timestamp",
        "timestamp": "timestamp",
        "year": "integer",
    }
    string_types = ["char", "varchar"]
    cmax = 65535


class MySQLTableConfig(TableConfig):
    """TableConfig Subclass for MySQL tables. This class will have all
    methods that are specific to MySQL, and handle the extracting and
    transforming of the data.

    Attributes
    ----------
    (see TableConfig)
    connection_kwargs : dict
        Dict created from environment variables that can be passed
        into MySQLdb or as an environment
    """

    database_type = "mysql"
    avro_type_map = {
        "string": pymysql.connections.TEXT_TYPES.union(
            {
                ft.DATE,
                ft.TIME,
                ft.TIMESTAMP,
                ft.JSON,
                ft.DATETIME,
                ft.ENUM,
                ft.YEAR,
                ft.NULL,
            }
        ),
        "int": {},  # prefer long to int
        "long": {ft.SHORT, ft.INT24, ft.TINY, ft.LONG, ft.LONGLONG},
        "double": {ft.DOUBLE, ft.FLOAT},
        "boolean": set(),
        "decimal": {ft.DECIMAL, ft.NEWDECIMAL},
    }

    @staticmethod
    def get_non_null_datetime_converters():
        """
        Build a substitute for mysql DT conversions which
        can handle "zero values" like '0000-00-00' - these
        are normally are converted to python None as they're
        invalid Datetimes, even if the column is NOT NULL.

        Note: converters are not called at all if a literal
        None is received - it short circuits. See:
        pymysql.connections.MySQLResult._read_row_from_packet

        Returns
        -------
        a dictionary of substitute timestamp conversions
        to pass to the `connect` call.

        """

        def convert_or_default(f, default):
            def _converter(x):
                out = f(x)
                # As of v0.8 pymysql casts "broken" datetimes to strings instead of None
                if out is not None and not isinstance(out, str):
                    return out
                else:
                    return default

            return _converter

        return {
            ft.DATE: convert_or_default(
                f=pymysql.converters.convert_date, default=datetime.date.min
            ),
            ft.DATETIME: convert_or_default(
                f=pymysql.converters.convert_datetime, default=datetime.datetime.min,
            ),
            ft.TIMESTAMP: convert_or_default(
                f=pymysql.converters.convert_mysql_timestamp,
                default=datetime.datetime.min,
            ),
        }

    @property
    def connection_vars(self):
        return {
            "host": self.db_host,
            "user": self.db_user,
            "password": self.db_password,
            "port": self.db_port,
            "db": self.db_name,
            "charset": "utf8",
        }

    def get_sql_description(self, sql):
        with closing(pymysql.connect(**self.connection_vars)) as conn:
            with closing(conn.cursor(pymysql.cursors.SSDictCursor)) as cursor:
                cursor.execute(sql + " LIMIT 1")
                return cursor.description

    def _mysql_to_redshift_type(self, input_type):
        # Note other non-implemented MySQL 8 types:
        #  - Spatial Data Types
        #  - JSON data type
        inp = input_type.lower()

        # Note: for enums, the type name is the same as the column name
        if inp in self.type_map:
            return self.type_map[inp]

        unsigned = "unsigned" in inp
        m = re.search(r"\((\d+)\)", inp)
        if m is not None:
            prec = int(m.group(1))
        else:
            prec = None

        inp = inp.replace("integer", "int")
        type_name = inp.split("(")[0].rstrip()

        # Numeric Types
        if type_name in MysqlTypes.numeric_types_always_promote:
            return MysqlTypes.numeric_type_promotions[type_name]
        elif type_name in MysqlTypes.numeric_type_promotions:
            if not unsigned:
                if type_name == "int":
                    return "integer"
                else:
                    return type_name
            else:
                return MysqlTypes.numeric_type_promotions[type_name]
        elif type_name in MysqlTypes.fixed_types:
            return inp.replace("unsigned", "").rstrip()
        elif type_name == "bit":
            if prec == 1:
                return "boolean"
            else:
                return "varchar({})".format(prec)

        # Date and Time Types
        # don't need precision or scale
        if type_name in MysqlTypes.date_and_time_types.keys():
            return MysqlTypes.date_and_time_types[type_name]

        # String Types
        if type_name in MysqlTypes.string_types:
            out_prec = min(MysqlTypes.cmax, 4 * prec)
            return "varchar({})".format(out_prec)

        # Sensible Defaults
        return "varchar({})".format(MysqlTypes.cmax)

    def _load_new_index_value(self):
        query = "SELECT MAX(`{}`) AS index_value FROM `{}`;".format(
            self.index_column, self.source_table_name
        )
        return self.query_fetchone(query)["index_value"]

    def query_to_redshift_create_table(self, sql, table_name):
        if self.schema_file:
            query = load_query(self.schema_file, CONFIG_DIR)
            create_table = query.rstrip("; \n")
            create_table += self.create_table_keys()
            return create_table
        else:
            desc = self.get_sql_description(sql)
            create_table = """CREATE TABLE "{}"."{}" (\n    """.format(
                self.destination_schema_name, table_name
            )
            field_strs = []
            for (name, type_code, _, _, precision, scale, null_ok,) in desc:
                # Note: mysql overreports this number by up to 4 places, which
                # should't cause problems
                size_str = "({}".format(precision) if precision else ""
                size_str += ",{}".format(scale) if scale else ""
                size_str += ")" if size_str else ""

                sql_type = "{}{}".format(
                    MysqlTypes.mysql_type_to_name[type_code], size_str
                )
                red_type = self._mysql_to_redshift_type(sql_type)
                field_strs.append(
                    '"{name}" {type}{null_ok}'.format(
                        name=name,
                        type=red_type,
                        null_ok="" if null_ok else " NOT NULL",
                    )
                )

            create_table += "\n  , ".join(field_strs)
            create_table += "\n)\n"
            create_table += self.create_table_keys()
            return create_table

    def query(self, sql):
        """Handles the opening and closing of connections for querying the
        source MySQL database

        Parameters
        ----------
        sql : str
            preformatted SQL query

        Returns
        -------
        list of dicts (generator)
            Returns the full database result. If a query returns no
            results, this returns an empty list.
        """

        self.logger.debug("Running query: %s", sql)

        converters = pymysql.converters.conversions
        if self.not_null_date:
            converters = converters.copy()
            converters.update(self.get_non_null_datetime_converters())

        with closing(pymysql.connect(conv=converters, **self.connection_vars)) as conn:
            with closing(conn.cursor(pymysql.cursors.SSDictCursor)) as cursor:
                cursor.execute(sql)
                for dict_row in cursor:
                    yield dict_row

    def _get_query_sql(self):
        if self.query_file:
            return self.get_query_from_file()

        if not self.pks:
            self.pks = [
                c["Column_name"]
                for c in self.query(
                    """
                SHOW KEYS FROM {} WHERE Key_name = 'PRIMARY'
                """.format(
                        self.source_table_name
                    )
                )
            ]

        cols = [
            c["column_name"]
            for c in self.query(
                """
            SELECT column_name
            FROM information_schema.columns
            WHERE table_schema = DATABASE()
                AND table_name='{}'
                AND column_name NOT IN ('{}')
            ORDER BY ordinal_position
            """.format(
                    self.source_table_name, "','".join(self.columns_to_drop)
                )
            )
        ]

        return "SELECT\n    `{}`\nFROM `{}`".format(
            "`\n  , `".join(cols), self.source_table_name
        )



import logging
from contextlib import contextmanager

import psycopg2

from druzhba.config import RedshiftConfig

logger = logging.getLogger("druzhba.redshift")


class Redshift(object):
    """Mixin for Redshift connection configs"""

    def __init__(self, destination_config):
        self.config = destination_config

    @property
    def iam_copy_role(self):
        return self.config.iam_copy_role

    @property
    def s3_config(self):
        return self.config.s3_config

    @contextmanager
    def connection(self):
        redshift_kwargs = self.config.connection_params

        if self.config.redshift_cert_path:
            redshift_kwargs.update(
                {"sslmode": "verify-ca", "sslrootcert": self.config.redshift_cert_path}
            )

        connection = psycopg2.connect(**redshift_kwargs)
        connection.set_client_encoding("utf-8")
        connection.autocommit = True
        try:
            yield connection
        finally:
            connection.close()

    @contextmanager
    def cursor(self, cursor_factory=None):
        with self.connection() as connection:
            cursor = connection.cursor(cursor_factory=cursor_factory)
            try:
                yield cursor
            finally:
                cursor.close()


_redshift = None


def get_redshift():
    return _redshift


def init_redshift(destination_config):
    # TODO: Replace with singleton pattern
    global _redshift  # pylint: disable=global-statement
    _redshift = Redshift(RedshiftConfig(destination_config))
    return _redshift


def generate_copy_query(table_to_copy, copy_target_url, iam_copy_role, manifest_mode):
    query = """
        COPY "{table_to_copy}" FROM '{s3_path}'
        CREDENTIALS 'aws_iam_role={iam_copy_role}'
        {manifest}
        FORMAT AS AVRO 'auto'
        EXPLICIT_IDS ACCEPTINVCHARS TRUNCATECOLUMNS
        COMPUPDATE OFF STATUPDATE OFF;
        """.format(
        table_to_copy=table_to_copy,
        s3_path=copy_target_url,
        iam_copy_role=iam_copy_role,
        manifest="MANIFEST" if manifest_mode else "",
    )
    return query


def generate_rename_query(current_table_name, renamed_table_name):
    return 'ALTER TABLE "{current_table_name}" RENAME TO "{renamed_table_name}";'.format(
        current_table_name=current_table_name, renamed_table_name=renamed_table_name,
    )


def generate_count_query(table):
    return "SELECT COUNT(*) FROM {};".format(table)


def generate_insert_all_query(table_to_select_from, table_to_insert_into):
    return 'INSERT INTO "{table_to_insert_into}" SELECT * FROM "{table_to_select_from}";'.format(
        table_to_select_from=table_to_select_from,
        table_to_insert_into=table_to_insert_into,
    )


def generate_create_table_like_query(new_table_name, table_to_copy):
    return 'CREATE TABLE "{}" (LIKE "{}");'.format(new_table_name, table_to_copy)


def generate_drop_query(table):
    return 'DROP TABLE "{}";'.format(table)


def generate_drop_exists_query(table):
    return 'DROP TABLE IF EXISTS "{}";'.format(table)


def generate_lock_query(table):
    return 'LOCK TABLE "{}";'.format(table)


def create_index_table(index_schema, index_table):
    logger.info(
        "Checking for existence of index table %s.%s", index_schema, index_table
    )
    with get_redshift().cursor() as cur:
        cur.execute(
            """SELECT COUNT(*) = 1
               FROM pg_catalog.pg_class c
               JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace
               WHERE n.nspname = %s
                 AND c.relname = %s
                 AND c.relkind = 'r'    -- only tables
            """,
            (index_schema, index_table),
        )
        # TODO: check format of this table that it's correct maybe?
        if not cur.fetchone()[0]:
            logger.warning(
                "Index table %s.%s does not exist, creating", index_schema, index_table
            )
            cur.execute(
                f"""
                CREATE TABLE {index_schema}.{index_table} (
                    datastore_name VARCHAR(127) NOT NULL,
                    database_name  VARCHAR(127) NOT NULL,
                    table_name     VARCHAR(127) NOT NULL,
                    index_value    VARCHAR(256) NOT NULL,
                    created_ts     TIMESTAMP DEFAULT getdate()
                )
                DISTSTYLE even
                SORTKEY(created_ts)
                ;
            """
            )
            logger.info("Index table %s.%s created", index_schema, index_table)



from contextlib import closing

import psycopg2
import psycopg2.extensions
import psycopg2.extras

from druzhba.config import CONFIG_DIR
from druzhba.table import TableConfig, load_query

psycopg2.extensions.register_type(psycopg2.extensions.UNICODE)
psycopg2.extensions.register_type(psycopg2.extensions.UNICODEARRAY)

ENCODING = "UTF8"


class PostgreSQLTableConfig(TableConfig):
    """TableConfig Subclass for PostgreSQL tables. This class will have all
    methods that are specific to Postgres, and handle the extracting and
    transforming of the data.

    Attributes
    ----------
    see TableConfig
    """

    database_type = "postgresql"
    avro_type_map = {
        "string": {
            "xml",
            "char",
            "text",
            "bytea",
            "name",
            "json",
            "jsonb",
            "varchar",
            "timestamp",
            "date",
            "time",
            "timestampz",
            "citext"
        },
        "int": {},  # prefer long to int
        "long": {"int2", "int4", "oid", "int8", "serial8"},
        "double": {"float4", "float8"},
        "boolean": {"bool"},
        "decimal": {"decimal", "numeric", "money"},
    }
    _pg_types = None

    def __init__(self, *args, **kwargs):
        super(PostgreSQLTableConfig, self).__init__(*args, **kwargs)
        type_map_defaults = {
            "text": "varchar(max)",
            "citext": "varchar(max)",
            "jsonb": "varchar(max)",
            "json": "varchar(max)",
            "array": "varchar(max)",
            "uuid": "char(36)",
        }

        type_map_defaults.update(self.type_map)
        self.type_map = type_map_defaults

    @property
    def connection_vars(self):
        return {
            "database": self.db_name,
            "host": self.db_host,
            "port": self.db_port,
            "user": self.db_user,
            "password": self.db_password,
        }

    @property
    def pg_types(self):
        if self._pg_types is None:
            with closing(psycopg2.connect(**self.connection_vars)) as conn:
                conn.set_client_encoding(ENCODING)
                with closing(
                    conn.cursor(
                        cursor_factory=psycopg2.extras.DictCursor, name="druzhba"
                    )
                ) as cursor:
                    cursor.execute("select oid, typname from pg_type")
                    self._pg_types = dict(
                        [(t["oid"], t["typname"]) for t in cursor.fetchall()]
                    )
        return self._pg_types

    def _get_column_is_nullable(self):
        is_nullable_query = """

        SELECT column_name, (is_nullable = 'YES') AS is_nullable
        FROM information_schema.columns
        WHERE table_name = '{0}';

        """.format(
            self.source_table_name
        )

        return {
            row["column_name"]: row["is_nullable"]
            for row in self.query(is_nullable_query)
        }

    def _load_new_index_value(self):
        query = 'SELECT MAX("{}") AS index_value FROM "{}";'.format(
            self.index_column, self.source_table_name
        )
        return self.query_fetchone(query)["index_value"]

    def get_sql_description(self, sql):
        if self.query_file is None:
            column_is_nullable = self._get_column_is_nullable()
        else:
            column_is_nullable = {}

        with closing(psycopg2.connect(**self.connection_vars)) as conn:
            conn.set_client_encoding(ENCODING)
            with closing(
                conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
            ) as cursor:
                cursor.execute(sql.rstrip("; \n") + " LIMIT 1")
                return (
                    (
                        col.name,
                        self.type_map.get(
                            self.pg_types[col.type_code], self.pg_types[col.type_code],
                        ),
                        None,
                        col.internal_size,
                        col.precision,
                        col.scale,
                        column_is_nullable.get(col.name),
                    )
                    for col in cursor.description
                )

    def query_to_redshift_create_table(self, sql, table_name):
        if self.schema_file:
            create_table = load_query(self.schema_file, CONFIG_DIR).rstrip("; \n\t")
            create_table += self.create_table_keys()
            return create_table
        else:
            if self.query_file is not None:
                self.logger.warning(
                    (
                        "Cannot obtain `null_ok` attribute for columns in postgres "
                        'source necessary to create target table "%s", assuming '
                        "that all columns should be NOT NULL. Please manually "
                        "rebuild the target table if some columns should be "
                        "nullable."
                    ),
                    table_name,
                )

            desc = self.get_sql_description(sql)
            create_table = """CREATE TABLE "{}"."{}" (\n    """.format(
                self.destination_schema_name, table_name
            )
            field_strs = []
            for (name, type_code, _, internal_size, precision, scale, null_ok,) in desc:
                size_str = "({}".format(precision) if precision else ""
                size_str += ",{}".format(scale) if scale else ""
                size_str += ")" if size_str else ""
                if self.type_map.get(type_code, type_code) == "varchar":
                    length = internal_size if internal_size > 0 else "max"
                    size_str = "({})".format(length)

                red_type = "{}{}".format(
                    self.type_map.get(type_code, type_code), size_str
                )
                field_strs.append(
                    '"{name}" {type}{null_ok}'.format(
                        name=name,
                        type=red_type,
                        null_ok="" if null_ok else " NOT NULL",
                    )
                )

            create_table += "\n  , ".join(field_strs)
            create_table += "\n)\n"
            create_table += self.create_table_keys()
            return create_table

    def query(self, sql):
        """Handles the opening and closing of connections for querying the
        source database

        Parameters
        ----------
        sql : str
            preformatted SQL query

        Returns
        -------
        list of dicts (generator)
            Returns the full database result. If a query returns no
            results, this returns an empty list.
        """

        self.logger.debug("Running query: %s", sql)

        cursor_name = "{}_{}".format(self.source_table_name, self.date_key)
        with closing(psycopg2.connect(**self.connection_vars)) as conn:
            conn.set_client_encoding(ENCODING)

            with conn.cursor(
                cursor_factory=psycopg2.extras.RealDictCursor, name=cursor_name
            ) as cursor:
                cursor.execute(sql)
                for dict_row in cursor:
                    yield dict_row

    def _get_query_sql(self):
        if self.query_file:
            return self.get_query_from_file()

        if not self.pks:
            self.pks = [
                c["attname"]
                for c in self.query(
                    """
                SELECT a.attname
                FROM pg_index i
                    JOIN pg_attribute a ON a.attrelid = i.indrelid
                                        AND a.attnum = ANY(i.indkey)
                WHERE i.indrelid = '{}'::regclass
                    AND i.indisprimary;
                """.format(
                        self.source_table_name
                    )
                )
            ]

        columns = [
            '"{}"'.format(c["column_name"])
            for c in self.query(
                """
            SELECT column_name
            FROM information_schema.columns
            WHERE table_schema = CURRENT_SCHEMA
                AND "table_name"= '{}'
                AND column_name NOT IN ('{}')
            ORDER BY ordinal_position
            """.format(
                    self.source_table_name, "','".join(self.columns_to_drop)
                )
            )
        ]

        return 'SELECT\n    {}\nFROM "{}"'.format(
            "\n  , ".join(columns), self.source_table_name
        )



import os
import re

import yaml

CONFIG_DIR = os.getenv("DRUZHBA_CONFIG_DIR")


class ConnectionConfig(object):
    # Note: pymysq defaults port to 3306 if 0 supplied
    #       pyscopg2 defaults port to 5432 if omitted from kwargs
    #       pymssql defaults to port 1433 if not provided
    def __init__(self, config):
        self.host = config.get("host")
        self.port = config.get("port")
        self.user = config.get("user")
        self.password = config.get("password")
        self.database = config.get("database")
        self.url = config.get("url")

    @property
    def params(self):
        if self.url:
            return {"dsn": self.url}
        elif self.host and self.user:
            return {
                "host": self.host,
                "port": self.port,
                "database": self.database,
                "user": self.user,
                "password": self.password,
            }
        else:
            raise ValueError("Required connection parameters not set")


class S3Config(object):
    def __init__(self, s3_config):
        self.bucket = s3_config.get("bucket", "").replace("s3://", "")
        self.prefix = s3_config.get("prefix", "")


class RedshiftConfig(object):
    """Either a URL or a host/port/database/user may be used."""

    def __init__(self, destination_config):
        self.iam_copy_role = destination_config.get("iam_copy_role")
        self.redshift_cert_path = destination_config.get("redshift_cert_path")
        self.s3_config = S3Config(destination_config.get("s3"))
        self.connection_config = ConnectionConfig(destination_config["connection"])

    @property
    def connection_params(self):
        return self.connection_config.params


def load_destination_config(config_dir):
    return load_config_file(os.path.join(config_dir, "_pipeline.yaml"))


class SourceConfig(object):
    pass


def load_config_file(filename):
    """Loads a Druzhba config file and process env var substitution

    Druzhba uses YAML configuration files for both sources and destination
    databases. The YAML is post processed to perform environment variable
    substitution. Patterns of the format ${ENV_VAR} are substituted with the
    content of the the ENV_VAR environment variable.

    If ENV_VAR is not set in the environment then it is replaced with an empty
    string and the missing_vars flag will be set. The caller can determine if
    this state represents an error.

    Note: substitutions only occur within string values.

    :param str filename: The path to the config file to be loaded
    :returns:
        - conf (:py:class:`dict`) - The parsed config dictionary
        - missing_vars (:py:class:`set`) - Set of missing environment variables.
            Empty set indicates no errors in loading the config environment
    """
    with open(filename, "r") as f:
        config = yaml.safe_load(f)
    return _parse_config(config)


def _parse_config(config):
    """Expected to be called within load_config_file
    A description of operation can be found there"""

    def _parse_element(config_element):
        if type(config_element) == dict:
            return _parse_dict_config(config_element)
        elif type(config_element) == list:
            return _parse_list_config(config_element)
        elif type(config_element) == str:
            return _parse_string_config(config_element)
        else:
            # Do not process elements of unknown type
            return config_element, set()

    def _parse_string_config(raw_string_value):
        output_string = raw_string_value

        missing = set()
        matches = re.findall(r"\${(\w+)}", raw_string_value)

        for match in matches:
            env_val = os.getenv(match)
            if env_val is not None:
                output_string = output_string.replace(f"${{{match}}}", env_val)
            else:
                output_string = output_string.replace(f"${{{match}}}", "")
                missing.add(match)

        return output_string, missing

    def _parse_dict_config(config):
        subbed_config = {}
        missing = set()
        for key in config:
            new_val, new_errors = _parse_element(config[key])
            subbed_config[key] = new_val
            missing.update(new_errors)
        return subbed_config, missing

    def _parse_list_config(config):
        subbed_config = []
        missing = set()
        for elem in config:
            new_val, new_errors = _parse_element(elem)
            subbed_config.append(new_val)
            missing.update(new_errors)
        return subbed_config, missing

    return _parse_element(config)



import os
from collections import namedtuple
from urllib.parse import urlparse

import psycopg2
import pymssql
import pymysql

from druzhba.mssql import MSSQLTableConfig
from druzhba.mysql import MySQLTableConfig
from druzhba.postgres import PostgreSQLTableConfig

ConnectionParams = namedtuple(
    "ConnectionParams", ["name", "host", "port", "user", "password"]
)


class DatabaseConfig(object):
    """Defines the configuration of a database.

    Should contain everything necessary to connect to a database.

    If neither 'connection_string' or 'connection_string_env' is given,
    connection parameters will be retrieved from the envar:

        '{DATABASE_ALIAS}_DATABASE_URL'

    Parameters
    ----------
    database_alias : str
        Colloquial name of the source database. Should match associated
        config files & env vars
    database_type : {'mysql', 'postgres'}
        Determines the engine of the source database
    connection_string : str, optional
        database connection string
    connection_string_env : str, optional
        name of database connection environment variable.
    object_schema_name : str, optional
        mssql object schema name
    db_template_data : dict, optional
        parameters to interpolate into SQL queries, if this is
        a SQL-defined table.

    Attributes
    ----------
    database_alias : str
        set by database_alias parameter
    database_type : str
        set by database_type parameter
    db_errors
        python namespace for appropriate DBAPI errors
    """

    def __init__(
        self,
        database_alias,
        database_type,
        connection_string=None,
        connection_string_env=None,
        object_schema_name=None,
        db_template_data=None,
    ):
        self.database_alias = database_alias
        self.database_type = database_type

        self._connection_string = connection_string
        self._connection_string_env = connection_string_env
        self._object_schema_name = object_schema_name
        self._db_template_data = db_template_data

        if self.database_type == "mysql":
            self._table_conf_cls = MySQLTableConfig
            self.db_errors = pymysql.err
        elif self.database_type == "postgres":
            self._table_conf_cls = PostgreSQLTableConfig
            self.db_errors = psycopg2
        elif self.database_type == "mssql":
            self._table_conf_cls = MSSQLTableConfig
            self.db_errors = pymssql
        else:
            msg = "Unknown database type {}".format(self.database_type)
            raise ValueError(msg)

    def get_table_config(self, table_params, index_schema, index_table):
        return self._table_conf_cls(
            self.database_alias,
            self.get_connection_params(),
            db_template_data=self._db_template_data,
            index_schema=index_schema,
            index_table=index_table,
            **table_params
        )

    def get_connection_params(self):
        if self._connection_string:
            db_url = self._connection_string
        else:
            if self._connection_string_env:
                env_path = self._connection_string_env
            else:
                env_path = self.database_alias.upper() + "_DATABASE_URL"
            db_url = os.getenv(env_path)
            if db_url is None:
                raise RuntimeError(
                    "Environment variable {} was not set".format(env_path)
                )

        parsed = urlparse(db_url)

        return ConnectionParams(
            name=self._object_schema_name or parsed.path.lstrip("/"),
            host=parsed.hostname,
            port=parsed.port,
            user=parsed.username,
            password=parsed.password,
        )



import argparse
import copy
import datetime
import logging
import sys
import time
import traceback
from multiprocessing import cpu_count
from multiprocessing.dummy import Pool

import psycopg2
import yaml
from botocore.vendored.requests.exceptions import SSLError

from druzhba.config import CONFIG_DIR, load_destination_config
from druzhba.db import DatabaseConfig
from druzhba.monitoring import DefaultMonitoringProvider, configure_logging
from druzhba.redshift import create_index_table, init_redshift
from druzhba.table import (
    ConfigurationError,
    InvalidSchemaError,
    MigrationError,
    TableConfig,
)

logger = logging.getLogger("druzhba.main")
monitor = DefaultMonitoringProvider()


COMPILE_ONLY = None
PRINT_SQL_ONLY = None
VALIDATE_ONLY = None


def process_database(
    index_schema,
    index_table,
    db_alias,
    db_type,
    only_table_names,
    full_refresh=None,
    rebuild=None,
):
    logger.info("Beginning database %s", db_alias)
    try:
        with monitor.wrap("run-time", database=db_alias):
            _process_database(
                index_schema,
                index_table,
                db_alias,
                db_type,
                only_table_names,
                full_refresh,
                rebuild,
            )
        logger.info("Done with database %s", db_alias)
    except Exception as e:
        logger.exception("Fatal error in database %s, aborting", db_alias)
        raise e


def _process_database(
    index_schema,
    index_table,
    db_alias,
    db_type,
    only_table_names,
    full_refresh=None,
    rebuild=None,
):
    with open("{}/{}.yaml".format(CONFIG_DIR, db_alias), "r") as f:
        dbconfig = yaml.safe_load(f)
        db = DatabaseConfig(
            db_alias,
            db_type,
            connection_string=dbconfig.get("connection_string"),
            connection_string_env=dbconfig.get("connection_string_env"),
            object_schema_name=dbconfig.get("data", {}).get("object_schema_name"),
            db_template_data=dbconfig.get("data", {}),
        )

    tables_yaml = dbconfig["tables"]

    logger.info(
        "%s tables ready to import for database %s",
        len(tables_yaml),
        db.database_alias,
    )

    invalids = []
    retries_remaining = 5

    for table_yaml in tables_yaml:
        if not only_table_names and not table_yaml.get("enabled", True):
            continue

        source_table_name = table_yaml["source_table_name"]
        if only_table_names and source_table_name not in only_table_names:
            continue

        try:
            TableConfig.validate_yaml_configuration(table_yaml)
            logger.info("Validated: %s / %s", db.database_alias, source_table_name)
        except ConfigurationError as e:
            logger.error(str(e))
            invalids.append(source_table_name)
            continue

        if VALIDATE_ONLY:
            continue

        table_params = copy.deepcopy(table_yaml)
        if rebuild:
            table_params["rebuild"] = True
            table_params["full_refresh"] = True
        elif full_refresh:
            table_params["full_refresh"] = True
        table = db.get_table_config(
            table_params, index_schema=index_schema, index_table=index_table
        )
        table.validate_runtime_configuration()

        if COMPILE_ONLY:
            print("---------------------------------------------")
            print(table.get_query_sql())
            print("---------------------------------------------\n\n\n")
            continue

        if PRINT_SQL_ONLY:
            select_query = table.get_query_sql()
            # Create statement introspects the source DB for a schema
            create_statement = table.query_to_redshift_create_table(
                select_query, table.destination_table_name
            )

            print("---------------------------------------------")
            print(create_statement)
            print("---------------------------------------------")
            print(select_query)
            print("---------------------------------------------\n\n\n")
            continue

        advance_to_next_table = False
        while not advance_to_next_table and retries_remaining > 0:
            try:
                with monitor.wrap("create-redshift-table", database=db_alias):
                    table.check_destination_table_status()

                with monitor.wrap(
                    "extract-table", database=db_alias, table=table.source_table_name
                ):
                    table.extract()

                with monitor.wrap(
                    "load-table", database=db_alias, table=table.source_table_name
                ):
                    table.load()

                advance_to_next_table = True

            except (InvalidSchemaError, MigrationError):
                logger.exception(
                    "Error preparing target table %s.%s",
                    table.destination_schema_name, table.destination_table_name
                )
                advance_to_next_table = True

            except (
                ValueError,
                db.db_errors.InternalError,
                db.db_errors.IntegrityError,
                db.db_errors.ProgrammingError,
                psycopg2.InternalError,
                psycopg2.IntegrityError,
                psycopg2.ProgrammingError,
                psycopg2.extensions.TransactionRollbackError,
                psycopg2.errors.FeatureNotSupported,  # pylint: disable=no-member
            ) as e:
                logger.warning(
                    "Unexpected error processing %s table %s: ```%s\n\n%s```",
                    table.database_alias,
                    table.source_table_name,
                    e,
                    "".join(traceback.format_exc()),
                )
                logger.info("Continuing")
                advance_to_next_table = True

            except (
                SSLError,
                db.db_errors.OperationalError,
                db.db_errors.DatabaseError,
            ) as e:
                retries_remaining -= 1
                if retries_remaining > 0:
                    logger.info(
                        "Disconnected while processing %s table %s with error... Retrying.",
                        table.database_alias,
                        table.source_table_name,
                    )
                    logger.info(e)
                    monitor.record_error("disconnect-error", db_alias=db_alias)
                    time.sleep((5.0 - retries_remaining) ** 2)
                else:
                    logger.error(
                        "Error processing %s table %s and out of retries: ```%s\n\n%s```",
                        table.database_alias,
                        table.source_table_name,
                        e,
                        "".join(traceback.format_exc()),
                    )
                    raise

            except (psycopg2.extensions.QueryCanceledError, Exception) as e:
                logger.error(
                    "Unexpected error processing %s table %s",
                    table.database_alias,
                    table.source_table_name,
                )
                raise

            logger.info(
                "Done with %s table %s", table.database_alias, table.source_table_name,
            )

    if len(invalids) > 0:
        raise RuntimeError(
            "Had invalid table configurations in {}: \n{}".format(
                db.database_alias, ",".join(invalids)
            )
        )


@monitor.timer("full-run-time")
def run(args):
    # pylint: disable=global-statement
    if args.tables and not args.database:
        msg = "--tables argument is not valid without --database argument"
        raise ValueError(msg)
    if args.full_refresh and not args.tables:
        msg = "--full-refresh argument is not valid without --table(s) argument"
        raise ValueError(msg)

    logger.info(
        "Detected %s CPUs available with %s threads requested. Using %s.",
        cpu_count(),
        args.num_processes or "unspecified",
        args.num_processes or cpu_count(),
    )

    global COMPILE_ONLY
    COMPILE_ONLY = args.compile_only

    global PRINT_SQL_ONLY
    PRINT_SQL_ONLY = args.print_sql_only

    global VALIDATE_ONLY
    VALIDATE_ONLY = args.validate_only

    destination_config, missing_vars = load_destination_config(CONFIG_DIR)
    if not COMPILE_ONLY and not PRINT_SQL_ONLY and not VALIDATE_ONLY:
        if missing_vars:
            logger.error(
                "Could not find required environment variable(s): %s",
                ", ".join(missing_vars),
            )
            sys.exit(1)

    index_schema = destination_config["index"]["schema"]
    index_table = destination_config["index"]["table"]

    init_redshift(destination_config)

    # Create the index table if it doesn't exist
    if not COMPILE_ONLY and not PRINT_SQL_ONLY and not VALIDATE_ONLY:
        create_index_table(index_schema, index_table)

    if args.database:
        dbs = [
            (
                index_schema,
                index_table,
                db["alias"],
                db["type"],
                args.tables,
                args.full_refresh,
                args.rebuild,
            )
            for db in destination_config["sources"]
            if db["alias"] == args.database
        ]
        if not dbs:
            msg = "Database {} not recognized in _databases.yml".format(args.database)
            raise ValueError(msg)
    else:
        dbs = [
            (
                index_schema,
                index_table,
                db["alias"],
                db["type"],
                args.tables,
                None,
                None,
            )
            for db in destination_config["sources"]
            if db.get("enabled", True)
        ]

    if args.num_processes == 1:
        for db in dbs:
            process_database(*db)
    else:
        # Preload _strptime to avoid a threading bug in cpython
        # See: https://mail.python.org/pipermail/python-list/2015-October/697689.html
        _ = datetime.datetime.strptime("2018-01-01 01:02:03", "%Y-%m-%d %H:%M:%S")
        with Pool(args.num_processes) as pool:
            results = pool.map_async(lambda db: process_database(*db), dbs)

            results.wait()
            if not results.successful():
                # Don't need to relog on failure, the process already logged
                sys.exit(2)

    if args.validate_only:
        logger.info("Validation complete")


def _get_parser():

    parser = argparse.ArgumentParser(description="Friendly data pipeline framework")
    parser.add_argument(
        "-ll", "--log-level", help="Name of a python log level eg DEBUG"
    )
    parser.add_argument(
        "-d",
        "-db",
        "--database",
        help="A single database to run."
        "\nWill override a database marked disabled in the db config file",
    )
    parser.add_argument(
        "-t",
        "--table",
        "--tables",
        help="List of tables to run separated by spaces. Must be run"
        " with --database",
        nargs="*",
        dest="tables",
    )
    parser.add_argument(
        "-np",
        "--num-processes",
        help="Number of parallel processes to spawn."
        "\nDefaults to number of CPUs (cores) available.",
        type=int,
    )
    parser.add_argument(
        "-co",
        "--compile-only",
        action="store_true",
        help="Will print generated queries to STDOUT but not execute anything.",
    )
    parser.add_argument(
        "-ps",
        "--print-sql-only",
        action="store_true",
        help="Will print generated CREATE and SELECT statements to STDOUT only.",
    )
    parser.add_argument(
        "-vo",
        "--validate-only",
        action="store_true",
        help="Will execute configuration checks only.",
    )
    parser.add_argument(
        "-f",
        "--full-refresh",
        help="Force a full refresh of the table. "
        "Must be run with --database and --table(s). ",
        action="store_true",
    )
    parser.add_argument(
        "-r",
        "--rebuild",
        help="Automatically recreate and full-refresh the table. "
        "Must be run with --database and --table(s). Only "
        "supported for tables Druzhba can build.",
        action="store_true",
    )
    return parser


def main():
    args = _get_parser().parse_args()

    configure_logging(args)
    logger.info("Running druzhba")

    run(args)

    logger.info("Shutting down")


if __name__ == "__main__":
    main()



import logging
import logging.config
import os
from contextlib import contextmanager
from enum import Enum, unique
from time import perf_counter

import sentry_sdk as sentry
import statsd
from sentry_sdk.integrations.logging import LoggingIntegration


def configure_logging(args):
    if args.log_level:
        log_level = args.log_level
    else:
        log_level = os.getenv("LOG_LEVEL", "INFO")

    settings = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "normal": {
                "format": "[%(asctime)s.%(msecs)03d] %(name)s [pid:%(process)s] - %(levelname)s - %(message)s",
                "datefmt": "%Y-%m-%d %H:%M:%S",
            }
        },
        "handlers": {
            "console": {
                "level": log_level,
                "class": "logging.StreamHandler",
                "formatter": "normal",
                "stream": "ext://sys.stdout",
            }
        },
        "loggers": {"druzhba": {"level": log_level, "handlers": ["console"]},},
    }
    logging.config.dictConfig(settings)


def init_sentry():
    dsn = os.getenv("SENTRY_DSN")

    if dsn is not None:
        sentry_logging = LoggingIntegration(
            level=logging.INFO, event_level=logging.WARNING
        )

        sentry.init(
            dsn=dsn,
            integrations=[sentry_logging],
            environment=os.getenv("SENTRY_ENVIRONMENT"),
            release=os.getenv("SENTRY_RELEASE"),
        )


class FakeStatsd(object):
    @contextmanager
    def timer(self, *args, **kwargs):
        # pylint: disable=unused-argument
        try:
            yield None
        finally:
            pass

    def timing(self, *args, **kwargs):
        pass

    def incr(self, *args, **kwargs):
        pass


def get_statsd_client():
    host = os.getenv("STATSD_HOST")
    port = os.getenv("STATSD_PORT")
    prefix = os.getenv("STATSD_PREFIX")
    if host and port:
        return statsd.StatsClient(host, port, prefix)
    else:
        return FakeStatsd()


@unique
class EventState(Enum):
    START = 1
    ERROR = 2
    COMPLETE = 3


class MonitoringProvider(object):
    """Provides real time monitoring of Druzhba

    The MonitoringProvider allows for custom monitoring of a running Druzhba
    process. The default implementation utilizes Statsd to track table and
    database extraction times and throughput.

    To implement a custom monitoring operation subclass MonitoringProvider and
    override the on_event method.
    """

    def __init__(self):
        pass

    def on_event(self, event, event_state, **kwargs):
        """Monitoring event handler

        Override to implement custom monitoring, logging, or stats collection

        At critical points within the life cycle of a Druzhba the application
        will invoke the on_event method of the current application's
        MonitoringProvider. It will be passed appropriate context based on when
        it was invoked by the application

        event: str
            The name of the event. See [events] in the docs for a list of event
            names

        event_state: EventState
            The state of the event started, completed, or errored

        db_alias: str (optional)
            The name of the database currently being processed

        table: str (optional)
            The name of the table currently being processed

        error: Exception (optional)
            The exception tha occurred within the relevant processing step. This
            exception is passed for logging or monitoring purposes and will be
            reraised within the Druzhba application after the on_event method
            is invoked.
        """

    @contextmanager
    def wrap(self, event, **kwargs):
        was_error = False

        self.on_event(event, EventState.START, **kwargs)
        started = perf_counter()

        try:
            yield None
        except Exception as ex:
            was_error = True
            et = perf_counter() - started
            kwargs.update({"et": et, "ex": ex})
            self.on_event(event, EventState.ERROR, **kwargs)
            raise
        finally:
            if not was_error:
                et = perf_counter() - started
                kwargs.update({"et": et})
                self.on_event(event, EventState.COMPLETE, **kwargs)

    def timer(self, event, **t_kwargs):
        def decorator(func):
            def wrapper(*args, **f_kwargs):
                with self.wrap(event, **t_kwargs):
                    out = func(*args, **f_kwargs)
                return out

            return wrapper

        return decorator

    def record_event(self, event, **kwargs):
        self.on_event(event, EventState.COMPLETE, **kwargs)

    def record_error(self, event, **kwargs):
        self.on_event(event, EventState.ERROR, **kwargs)


class DefaultMonitoringProvider(MonitoringProvider):
    """Implements monitoring for Druzhba application process

    The default monitoring includes a statsd integration that will, if
    STATSD_HOST and STATSD_PORT environment variables are defined, post run
    metadata to the appropriate statsd instance.  If you would like to implement
    custom monitoring and also include statsd behavior, subclass this class. If
    you would prefer to not include default statsd behavior subclass
    MonitoringProvider directly"""

    def __init__(self):
        super(DefaultMonitoringProvider).__init__()
        self.statsd_client = get_statsd_client()

    def on_event(self, event, event_state, **kwargs):
        db_alias = kwargs.get("db_alias")
        table = kwargs.get("table")
        elapsed_time = kwargs.get("et")

        if db_alias and table:
            full_event_name = f"druzhba.db.{event}.{db_alias}.{table}"
        elif db_alias:
            full_event_name = f"druzhba.db.{event}.{db_alias}"
        else:
            full_event_name = f"druzhba.db.{event}"

        if event_state == EventState.COMPLETE:
            if event in ["extract-table", "create-redshift-table", "load-table"]:
                statsd_client.incr(full_event_name)
            if event in ["run-time", "full-run-time"]:
                statsd_client.timing(full_event_name, elapsed_time)

        if event == "disconnect-error":
            statsd_client.incr(f"druzhba.db.{event}.{db_alias}")


statsd_client = get_statsd_client()
init_sentry()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2019 SeatGeek
import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

package_vars = {}
with open("druzhba/_version.py", "r") as f:
    exec(f.read(), package_vars)

tests_require = [
    "nose>=1.3.7",
    "mock>=2.0.0",
    "pylint==2.5.3",
]

setuptools.setup(
    name="druzhba",
    version=package_vars['__version__'],
    author="Seatgeek and Contributors",
    author_email="druzhba-maintainers@seatgeek.com",
    description="A friendly data pipeline framework",
    license="MIT License",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/seatgeek/druzhba",
    packages=setuptools.find_packages(exclude=['test']),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
    python_requires=">=3.6",
    setup_requires=[
        "cython>=0.29.7"  # TODO: find a way to not pre-install this (needed for mssql)
    ],
    # TODO: move specific DBs into extras
    install_requires=[
        "boto3>=1.10.34",
        "botocore>=1.13.35",
        "fastavro>=0.21.22,<0.22",
        "Jinja2>=2.10",
        "psycopg2-binary>=2.7.3.2",
        "pyaml>=17.10.0",
        "pymssql<3.0",  # TODO: replace with PyODBC
        "pymysql>=0.7.11",
        "sentry-sdk>=0.11.0,<0.14",
        "statsd>=3.3.0",
    ],
    extras_require={
        "dev": ["autoflake", "black", "isort", "sphinx", "recommonmark", "twine"],
        "test": tests_require
    },
    entry_points={"console_scripts": ["druzhba=druzhba.main:main"]},
    scripts=[],
    tests_require=tests_require
)
